# -*- coding: utf-8 -*-
"""
lists
~~~~~

Collections based on list interface.
"""
from __future__ import division, print_function, unicode_literals

import collections

import six

from .base import RedisCollection


class List(RedisCollection, collections.MutableSequence):
    """Mutable **sequence** collection aiming to have the same API as the
    standard sequence type, :class:`list`. See `list
    <http://docs.python.org/2/library/functions.html#list>`_ for
    further details. The Redis implementation is based on the
    `list <http://redis.io/commands#list>`_ type.

    .. warning::
        In comparing with original :class:`list` type, :class:`List` does not
        implement methods :func:`sort` and :func:`reverse`.

    .. note::
        Some operations, which are usually not used so often, can be more
        efficient than their "popular" equivalents. Some operations are
        shortened in their capabilities and can raise
        :exc:`NotImplementedError` for some inputs (e.g. most of the slicing
        functionality).
    """

    def __init__(self, *args, **kwargs):
        """
        :param data: Initial data.
        :type data: iterable
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str
        :param writeback: If ``True`` keep a local cache of changes for storing
                          modifications to mutable values. Changes will be
                          written to Redis after calling the ``sync`` method.
        :type key: bool

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.
        """
        writeback = kwargs.pop('writeback', False)
        super(List, self).__init__(*args, **kwargs)

        self.writeback = writeback
        self.cache = {}

    def _get_cache_index(self, index):
        return index if index >= 0 else self.__len__() + index

    def __len__(self):
        """Length of the sequence."""
        return self.redis.llen(self.key)

    def _data(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        values = redis.lrange(self.key, 0, -1)
        return (self._unpickle(v) for v in values)

    def __iter__(self):
        """Return an iterator over the sequence."""
        return (self.cache.get(i, x) for i, x in enumerate(self._data()))

    def __reversed__(self):
        """Returns iterator for the sequence in reversed order."""
        return reversed(list(self.__iter__()))

    def _recalc_slice(self, start, stop):
        """Slicing in Redis takes also the item at 'stop' index, so there is
        some recalculation to be done. Method returns tuple ``(start, stop)``
        where both values are recalculated to numbers in Redis terms.

        :param start: Index starting the range (in Python terms).
        :param stop: Index where the range ends (in Python terms).
        """
        start = start or 0
        if stop is None:
            stop = -1
        else:
            stop = stop - 1 if stop != 0 else stop
        return start, stop

    def _calc_overflow(self, size, index):
        """Index overflow detection. Returns :obj:`True` if *index* is out of
        range for the given *size*. Otherwise returns :obj:`False`.

        :param size: Size of the collection.
        :param index: Index to be examined.
        """
        return (index >= size) if (index >= 0) else (abs(index) > size)

    def _get_slice(self, index):
        """Helper for getting a slice."""
        assert isinstance(index, slice)

        # For slices without a step we can use the Redis range function
        def slice_trans(pipe):
            calc_start, calc_stop = self._recalc_slice(index.start, index.stop)
            if calc_start == index.stop:
                return []

            values = []
            redis_values = pipe.lrange(self.key, calc_start, calc_stop)
            for i, v in enumerate(redis_values, index.start or 0):
                values.append(self.cache.get(i, self._unpickle(v)))

            pipe.multi()
            return self._create_new(values, pipe=pipe)

        # Otherwise we'll need to pull the whole list and slice in Python
        def step_trans(pipe):
            values = []
            redis_values = pipe.lrange(self.key, 0, -1)
            for i, v in enumerate(redis_values):
                values.append(self.cache.get(i, self._unpickle(v)))

            values = values[index.start:index.stop:index.step]
            pipe.multi()
            return self._create_new(values, pipe=pipe)

        if index.step:
            return self._transaction(step_trans)
        return self._transaction(slice_trans)

    def __getitem__(self, index):
        """Returns item of sequence on *index*.
        Origin of indexes is 0. Accepts also slicing.

        .. note::
            Due to implementation on Redis side, ``l[index]`` is not very
            efficient operation. If possible, use :func:`get`. Slicing without
            steps is efficient. Steps are implemented only on Python side.
        """
        if isinstance(index, slice):
            return self._get_slice(index)

        if self.writeback:
            cache_index = self._get_cache_index(index)
            if cache_index in self.cache:
                return self.cache[cache_index]

        with self.redis.pipeline() as pipe:
            pipe.llen(self.key)
            pipe.lindex(self.key, index)
            size, pickled_value = pipe.execute()

        if self._calc_overflow(size, index):
            raise IndexError(index)

        value = self._unpickle(pickled_value)
        if self.writeback:
            self.cache[cache_index] = value
        return value

    def get(self, index, default=None):
        """Return the value for *index* if *index* is not out of range, else
        *default*. If *default* is not given, it defaults to :obj:`None`, so
        that this method never raises a :exc:`IndexError`.

        .. note::
            Due to implementation on Redis side, this method of retrieving
            items is more efficient than classic approach over using the
            :func:`__getitem__` protocol.
        """
        if self.writeback:
            cache_index = self._get_cache_index(index)
            if cache_index in self.cache:
                return self.cache[cache_index]

        pickled_value = self.redis.lindex(self.key, index)
        if pickled_value is None:
            value = default
        else:
            value = self._unpickle(pickled_value)

        if self.writeback:
            self.cache[cache_index] = value
        return value

    def _set_slice(self, index, value):
        """Helper for setting a slice."""
        assert isinstance(index, slice)

        if value:
            # assigning anything else than empty lists not supported
            raise NotImplementedError(self.not_impl_msg)

        self.__delitem__(index)

    def __setitem__(self, index, value):
        """Item of *index* is replaced by *value*.

        .. warning::
            Slicing is generally not supported. Only empty lists are accepted
            if the operation leads into trimming::

                l[2:] = []
                l[:2] = []
                l[:] = []
        """
        if isinstance(index, slice):
            return self._set_slice(index, value)

        def set_trans(pipe):
            size = pipe.llen(self.key)
            if self._calc_overflow(size, index):
                raise IndexError(index)
            pipe.multi()
            pipe.lset(self.key, index, self._pickle(value))

        self._transaction(set_trans)

        if self.writeback:
            cache_index = self._get_cache_index(index)
            self.cache[cache_index] = value

    def _del_slice(self, index):
        """Helper for deleting a slice."""
        assert isinstance(index, slice)

        begin = 0
        end = -1

        if index.step:
            # stepping not supported
            raise NotImplementedError(self.not_impl_msg)

        start, stop = self._recalc_slice(index.start, index.stop)

        if start == begin and stop == end:
            # trim from beginning to end
            self.clear()
            self.cache.clear()
            return

        with self.redis.pipeline() as pipe:
            if self.writeback:
                self._sync_helper(pipe)
            if start != begin and stop == end:
                # right trim
                pipe.ltrim(self.key, begin, start - 1)
            elif start == begin and stop != end:
                # left trim
                pipe.ltrim(self.key, stop + 1, end)
            else:
                # only trimming is supported
                raise NotImplementedError(self.not_impl_msg)
            pipe.execute()

    def __delitem__(self, index):
        """Item of *index* is deleted.

        .. warning::
            Slicing is generally not supported. Only empty lists are accepted
            if the operation leads into trimming::

                del l[2:]
                del l[:2]
                del l[:]
        """
        begin = 0
        end = -1

        # Calculate the cache index before the length changes
        if self.writeback:
            cache_index = self._get_cache_index(index)

        if isinstance(index, slice):
            return self._del_slice(index)

        if index == begin:
            self.redis.lpop(self.key)
        elif index == end:
            self.redis.rpop(self.key)
        else:
            raise NotImplementedError(self.not_impl_msg)

        # Removing an item from the list means all the other items after it
        # have to shift back one - reflect that in the cache
        if self.writeback:
            new_cache = {}
            for k, v in six.iteritems(self.cache):
                if k < cache_index:
                    new_cache[k] = v
                elif k == cache_index:
                    pass
                elif k > cache_index:
                    new_cache[k - 1] = v

            self.cache = new_cache

    def remove(self, value):
        """Remove the first occurence of *value*."""
        # If we're caching, we'll need to synchronize before removing.
        with self.redis.pipeline() as pipe:
            if self.writeback:
                self._sync_helper(pipe)
            pipe.lrem(self.key, 1, self._pickle(value))
            pipe.execute()

    def index(self, value, start=None, stop=None):
        """Returns index of the first occurence of *value*.

        If *start* or *stop* are provided, returns the smallest
        index such that ``s[index] == value`` and ``start <= index < stop``.
        """
        for k, v in six.iteritems(self.cache):
            if v == value:
                if start is not None and k < start:
                    continue
                if stop is not None and k >= stop:
                    break
                return k

        start, stop = self._recalc_slice(start, stop)
        values = self.redis.lrange(self.key, start, stop)

        for i, v in enumerate(self._unpickle(v) for v in values):
            if v == value:
                return i + start
        raise ValueError(value)

    def count(self, value):
        """Returns number of occurences of *value*.

        .. note::
            Implemented only on Python side.
        """
        ret = 0
        for k, v in enumerate(self._data()):
            v = self.cache.get(k, v)
            if v == value:
                ret += 1

        return ret

    def insert(self, index, value):
        """Insert *value* before *index*. Can only work with index == 0.
        """
        if index != 0:
            # Redis has no commands for *inserting* into a list by index.
            # LINSERT requires assumptions about contents of the list values.
            raise NotImplementedError(self.not_impl_msg)

        self.redis.lpush(self.key, self._pickle(value))

        if self.writeback:
            new_cache = {k + 1: v for k, v in six.iteritems(self.cache)}
            new_cache[0] = value
            self.cache = new_cache

    def append(self, value):
        """Insert *value* at end of list.
        """
        i = self.redis.rpush(self.key, self._pickle(value))

        if self.writeback:
            self.cache[i - 1] = value

    def _update(self, data, pipe=None):
        super(List, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        values = [self._pickle(x) for x in data]
        return redis.rpush(self.key, *values)

    def extend(self, values):
        """*values* are appended at the end of the list. Any iterable
        is accepted.
        """
        if isinstance(values, RedisCollection):
            # wrap into transaction
            def extend_trans(pipe):
                d = values._data(pipe=pipe)  # retrieve
                pipe.multi()
                return self._update(d, pipe=pipe)  # store
            new_len = self._transaction(extend_trans)
        else:
            new_len = self._update(values)

        if self.writeback:
            for k, v in enumerate(values, new_len - len(values)):
                self.cache[k] = v

    def pop(self, index=-1):
        """Item on *index* is removed and returned.

        .. warning::
            Only indexes ``0`` and ``-1`` (default) are supported, otherwise
            :exc:`NotImplementedError` is raised.
        """
        # Calculate the cache index before the length changes
        if self.writeback:
            cache_index = self._get_cache_index(index)

        if index == 0:
            value = self._unpickle(self.redis.lpop(self.key))
        elif index == -1:
            value = self._unpickle(self.redis.rpop(self.key))
        else:
            raise NotImplementedError(self.not_impl_msg)

        # Removing an item from the list means all the other items after it
        # have to shift back one - reflect that in the cache
        if self.writeback:
            new_cache = {}
            for k, v in six.iteritems(self.cache):
                if k < cache_index:
                    new_cache[k] = v
                elif k == cache_index:
                    value = v
                elif k > cache_index:
                    new_cache[k - 1] = v

            self.cache = new_cache

        return value

    def __add__(self, values):
        """Returns concatenation of the list and given iterable. New
        :class:`List` instance is returned.
        """
        def add_trans(pipe):
            if self.writeback:
                self._sync_helper(pipe)

            d1 = list(self._data(pipe=pipe))  # retrieve

            if isinstance(values, RedisCollection):
                d2 = list(values._data(pipe=pipe))  # retrieve
            else:
                d2 = list(values)

            pipe.multi()
            return self._create_new(d1 + d2, pipe=pipe)  # store
        return self._transaction(add_trans)

    def __radd__(self, values):
        return self.__add__(values)

    def __mul__(self, n):
        """Returns *n* copies of the list, concatenated. New :class:`List`
        instance is returned.
        """
        if not isinstance(n, int):
            raise TypeError('Cannot multiply sequence by non-int.')

        def mul_trans(pipe):
            if self.writeback:
                self._sync_helper(pipe)

            data = list(self._data(pipe=pipe))  # retrieve
            pipe.multi()
            return self._create_new(data * n, pipe=pipe)  # store
        return self._transaction(mul_trans)

    def __rmul__(self, n):
        return self.__mul__(n)

    def _repr_data(self, data):
        return repr(list(data))

    def __enter__(self):
        self.writeback = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.sync()

    def _sync_helper(self, pipe):
        for k, v in six.iteritems(self.cache):
            pipe.lset(self.key, k, self._pickle(v))

        self.cache = {}

    def sync(self):
        def sync_trans(pipe):
            pipe.multi()
            self._sync_helper(pipe)

        self._transaction(sync_trans)



# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

__title__ = 'redis-collections'
__version__ = '0.2.0'
__author__ = 'Honza Javorek'
__license__ = 'ISC'
__copyright__ = 'Copyright 2013-? Honza Javorek'


from .base import RedisCollection  # NOQA
from .dicts import DefaultDict, Dict, Counter  # NOQA
from .lists import List  # NOQA
from .sets import Set  # NOQA

__all__ = ['RedisCollection', 'Set', 'List', 'DefaultDict', 'Dict', 'Counter']



# -*- coding: utf-8 -*-
"""
sets
~~~~

Collections based on set interface.
"""
from __future__ import division, print_function, unicode_literals

import abc
import itertools
import collections

import six

from .base import RedisCollection, same_types


@six.add_metaclass(abc.ABCMeta)
class SetOperation(object):
    """Helper class for implementing standard set operations."""

    def __init__(self, s, update=False, flipped=False, return_cls=None):
        """
        :param s: :class:`collections.Set` instance.
        :param update: If :obj:`True`, operation is considered to be *update*.
                       That means it affects directly the *s* object and
                       returns the original object itself.
         :param flipped: Specifies whether the operation is in reversed mode,
                         where *s* is the right operand and *other* given to
                         :func:`__call__` is the left one. With this option
                         *update* must be :obj:`False` and only one other
                         operand is accepted in :func:`__call__`.
        :param return_cls: Class object specifying the type to be used for
                            result of the operation. If *update* or *flipped*
                            are :obj:`True`, this argument is ignored.
        """
        assert not (update and flipped)

        self.s = s
        self.update = update
        self.flipped = flipped
        self.return_cls = None if (update or flipped) else return_cls

    def _are_set_instances(self, *others):
        """Helper method deciding whether given *others* are instances
        of :class:`Set` (sub)class.

        :param others: Any objects.
        :rtype: boolean
        """
        test = lambda other: isinstance(other, Set)
        if len(others) == 1:
            return test(others[0])
        return all(map(test, others))

    def _to_set(self, c, pipe=None):
        if isinstance(c, RedisCollection):
            return set(c._data(pipe=pipe))
        return set(c)

    def _op(self, key, return_cls, others):
        """:func:`op` wrapper. Takes care of proper transaction
        handling and result instantiation.
        """
        if self.flipped:
            assert len(others) == 1
            left = others[0]
            right = [self.s]
        else:
            left = self.s
            right = others

        def trans(pipe):
            # retrieve
            data = self._to_set(left, pipe=pipe)
            other_sets = [self._to_set(o, pipe=pipe) for o in right]

            # operation
            elements = self.op(data, other_sets)
            pipe.multi()

            # store within the transaction
            return self.s._create_new(elements, key=key, cls=return_cls,
                                      pipe=pipe)
        return self.s._transaction(trans, key)

    @abc.abstractmethod
    def op(self, s, other_sets):
        """Implementation of the operation on standard :class:`set`.

        :param s: Data of the original collection as set (first operand).
        :type s: :class:`set`
        :param other_sets: Data of all the other collections participating
                           in this operation as sets (other operands).
        :type other_keys: iterable of :class:`frozenset` instances
        :rtype: resulting iterable
        """
        pass

    def _redisop(self, key, return_cls, other_keys):
        """:func:`redisop` wrapper. Takes care of proper transaction
        handling and result instantiation.
        """
        if self.flipped:
            assert len(other_keys) == 1
            left = other_keys[0]
            right = [self.s.key]
        else:
            left = self.s.key
            right = other_keys

        def trans(pipe):
            # operation
            elements = self.redisop(pipe, left, right)
            pipe.multi()

            # store within the transaction
            return self.s._create_new(elements, key=key, cls=return_cls,
                                      pipe=pipe)
        return self.s._transaction(trans, key, *other_keys)

    @abc.abstractmethod
    def redisop(self, pipe, key, other_keys):
        """Implementation of the operation in Redis. Results
        are returned to Python.

        :param pipe: Redis transaction pipe.
        :type pipe: :class:`redis.client.StrictPipeline`
        :param key: Redis key from the original collection (first operand).
        :type key: string
        :param other_keys: Redis keys of all the other collections
                           participating in this operation (other operands).
        :type other_keys: iterable of strings
        :rtype: resulting iterable
        """
        pass

    def _redisopstore(self, key, return_cls, other_keys):
        """:func:`redisopstore` wrapper. Takes care of proper transaction
        handling and result instantiation.
        """
        if self.flipped:
            assert len(other_keys) == 1
            left = other_keys[0]
            right = [self.s.key]
        else:
            left = self.s.key
            right = other_keys

        def trans(pipe):
            # operation & possible store (in self.redisopstore)
            new = self.s._create_new(key=key, cls=return_cls)
            self.redisopstore(pipe, new.key, left, right)
            return new
        return self.s._transaction(trans, key, *other_keys)

    @abc.abstractmethod
    def redisopstore(self, pipe, new_key, key, other_keys):
        """Implementation of the operation in Redis. Results
        are stored to another key within Redis.

        :param pipe: Redis transaction pipe.
        :type pipe: :class:`redis.client.StrictPipeline`
        :param new_key: Redis key of the new collection (destination).
        :type new_key: string
        :param key: Redis key from the original collection (first operand).
        :type key: string
        :param other_keys: Redis keys of all the other collections
                           participating in this operation (other operands).
        :type other_keys: iterable of strings
        :rtype: :obj:`None`
        """
        pass

    def __call__(self, *others):
        """Operation trigger.

        :param others: Iterable of one or more iterables, which are part
                       of this operation.
        """
        if self.flipped:
            # should return type of the left operand
            assert len(others) == 1
            return_cls = others[0].__class__
        elif self.update:
            # should return the original set
            return_cls = self.s.__class__
        else:
            # should return type of the left operand or type
            # specified in self.return_cls
            return_cls = self.return_cls or Set

        new_key = self.s.key if self.update else None

        if self._are_set_instances(*others):
            # all others are of Set type
            other_keys = [other.key for other in others]

            if issubclass(return_cls, self.s.__class__):
                # operation can be performed in Redis completely
                return self._redisopstore(new_key, return_cls, other_keys)
            else:
                # operation can be performed in Redis and returned to Python
                return self._redisop(new_key, return_cls, other_keys)

        # else do it in Python completely,
        # simulating the same operation on standard set
        return self._op(new_key, return_cls, others)


class SetDifference(SetOperation):

    def op(self, s, other_sets):
        if self.update:
            s.difference_update(*other_sets)
            return s
        return s.difference(*other_sets)

    def redisop(self, pipe, key, other_keys):
        return pipe.sdiff(key, *other_keys)

    def redisopstore(self, pipe, new_key, key, other_keys):
        pipe.multi()
        pipe.sdiffstore(new_key, key, *other_keys)


class SetIntersection(SetOperation):

    def op(self, s, other_sets):
        if self.update:
            s.intersection_update(*other_sets)
            return s
        return s.intersection(*other_sets)

    def redisop(self, pipe, key, other_keys):
        return pipe.sinter(key, *other_keys)

    def redisopstore(self, pipe, new_key, key, other_keys):
        pipe.multi()
        pipe.sinterstore(new_key, key, *other_keys)


class SetUnion(SetOperation):

    def op(self, s, other_sets):
        if self.update:
            s.update(*other_sets)
            return s
        return s.union(*other_sets)

    def redisop(self, pipe, key, other_keys):
        return pipe.sunion(key, *other_keys)

    def redisopstore(self, pipe, new_key, key, other_keys):
        pipe.multi()
        pipe.sunionstore(new_key, key, *other_keys)


class SetSymmetricDifference(SetOperation):

    def op(self, s, other_sets):
        if self.update:
            s.symmetric_difference_update(*other_sets)
            return s
        return s.symmetric_difference(*other_sets)

    def _simulate_redisop(self, pipe, key, other_key):
        diff1 = pipe.sdiff(key, other_key)
        diff2 = pipe.sdiff(other_key, key)
        return diff1 | diff2  # return still pickled

    def redisop(self, pipe, key, other_keys):
        other_key = other_keys[0]  # sym. diff. supports only one operand
        elements = self._simulate_redisop(pipe, key, other_key)
        return [self.s._unpickle(x) for x in elements]

    def redisopstore(self, pipe, new_key, key, other_keys):
        other_key = other_keys[0]  # sym. diff. supports only one operand
        elements = self._simulate_redisop(pipe, key, other_key)  # pickled
        pipe.multi()
        pipe.delete(new_key)
        pipe.sadd(new_key, *elements)  # store pickled elements


class Set(RedisCollection, collections.MutableSet):
    """Mutable **set** collection aiming to have the same API as the standard
    set type. See `set
    <http://docs.python.org/2/library/stdtypes.html#set>`_ for
    further details. The Redis implementation is based on the
    `set <http://redis.io/commands#set>`_ type.
    """

    _same_types = (collections.Set,)

    def __init__(self, *args, **kwargs):
        """
        :param data: Initial data.
        :type data: iterable
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.
        """
        super(Set, self).__init__(*args, **kwargs)

    def __len__(self):
        """Return cardinality of the set."""
        return self.redis.scard(self.key)

    def _data(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return (self._unpickle(v) for v in redis.smembers(self.key))

    def __iter__(self):
        """Return an iterator over elements of the set."""
        return self._data()

    def __contains__(self, elem):
        """Test for membership of *elem* in the set."""
        return self.redis.sismember(self.key, self._pickle(elem))

    def add(self, elem):
        """Add element *elem* to the set. Returns :obj:`False` if
        *elem* was already present in the set.

        :rtype: boolean

        .. warning::
            Original :func:`add` in :class:`set` returns no value.
        """
        return bool(self.redis.sadd(self.key, self._pickle(elem)))

    def discard(self, elem):
        """Remove element *elem* from the set if it is present."""
        self.redis.srem(self.key, self._pickle(elem))

    def remove(self, elem):
        """Remove element *elem* from the set. Raises :exc:`KeyError` if elem
        is not contained in the set.
        """
        removed_count = self.redis.srem(self.key, self._pickle(elem))
        if not removed_count:
            raise KeyError(elem)

    def pop(self):
        """Remove and return an arbitrary element from the set.
        Raises :exc:`KeyError` if the set is empty.
        """
        with self.redis.pipeline() as pipe:
            pipe.scard(self.key)
            pipe.spop(self.key)
            size, elem = pipe.execute()

        if not size:
            raise KeyError
        return self._unpickle(elem)

    def random_sample(self, k=1):
        """Return a *k* length list of unique elements chosen from the set.
        Elements are not removed. Similar to :func:`random.sample` function
        from standard library.

        :param k: Size of the sample, defaults to 1.
        :rtype: :class:`list`

        .. note::
            Argument *k* is supported only for Redis of version 2.6 and higher.
        """
        if k < 1:
            return []
        if k == 1:
            elements = [self.redis.srandmember(self.key)]
        else:
            elements = self.redis.srandmember(self.key, k)
        return [self._unpickle(x) for x in elements]

    def difference(self, *others, **kwargs):
        """Return a new set with elements in the set that are
        not in the *others*.

        :param others: Iterables, each one as a single positional argument.
        :param return_cls: Keyword argument, type of result, defaults to
                            the same type as collection (:class:`Set`,
                            if not inherited).
        :rtype: :class:`Set` or collection of type specified in
                *return_cls* argument

        .. note::
            If all *others* are :class:`Set` instances, operation
            is performed completely in Redis. If *return_cls* is provided,
            operation is still performed in Redis, but results are sent
            back to Python and returned with corresponding type. All other
            combinations are performed in Python and results are sent
            to Redis. See examples::

                s1 = Set([1, 2])
                s2 = Set([2, 3])
                s3 = set([2, 3])  # built-in set

                # Redis (whole operation)
                s1.difference(s2, s2, s2)  # = Set

                # Python (operation) → Redis (new key with Set)
                s1.difference(s3)  # = Set

                # Python (operation) → Redis (new key with Set)
                s1.difference(s2, s3, s2)  # = Set

                # Redis (operation) → Python (type conversion)
                s1.difference(s2, return_cls=set)  # = set

                # Redis (operation) → Python (type conversion)
                s1.difference(s2, return_cls=list)  # = list

                # Redis (operation) → Python → Redis (new key with List)
                s1.difference(s2, return_cls=List)  # = List
        """
        return_cls = kwargs.get('return_cls', type(self))
        op = SetDifference(self, return_cls=return_cls)
        return op(*others)

    @same_types
    def __sub__(self, other):
        """Return a new set with elements in the set that are
        not in the *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: type of the first operand

        .. note::
            If *other* is instance of :class:`Set`, operation
            is performed completely in Redis. Otherwise it's performed
            in Python and results are sent to Redis.
        """
        return self.difference(other)

    @same_types
    def __rsub__(self, other):
        op = SetDifference(self, flipped=True)
        return op(other)

    def difference_update(self, *others):
        """Update the set, removing elements found in *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            If all *others* are :class:`Set` instances, operation
            is performed completely in Redis. Otherwise it's performed
            in Python and results are sent to Redis. See examples::

                s1 = Set([1, 2])
                s2 = Set([2, 3])
                s3 = set([2, 3])  # built-in set

                # Redis (whole operation)
                s1.difference_update(s2, s2)  # = None

                # Python (operation) → Redis (update)
                s1.difference(s3)  # = None

                # Python (operation) → Redis (update)
                s1.difference(s2, s3, s2)  # = None
        """
        op = SetDifference(self, update=True)
        op(*others)

    @same_types
    def __isub__(self, other):
        """Update the set, removing elements found in *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: None

        .. note::
            If *other* is instance of :class:`Set`, operation
            is performed completely in Redis. Otherwise it's performed
            in Python and results are sent to Redis.
        """
        op = SetDifference(self, update=True)
        return op(other)

    def intersection(self, *others, **kwargs):
        """Return a new set with elements common to the set and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :param return_cls: Keyword argument, type of result, defaults to
                            the same type as collection (:class:`Set`,
                            if not inherited).
        :rtype: :class:`Set` or collection of type specified in
                *return_cls* argument

        .. note::
            The same behavior as at :func:`difference` applies.
        """
        return_cls = kwargs.get('return_cls', type(self))
        op = SetIntersection(self, return_cls=return_cls)
        return op(*others)

    @same_types
    def __and__(self, other):
        """Return a new set with elements common to the set and the *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: type of the first operand

        .. note::
            The same behavior as at :func:`__sub__` applies.
        """
        return self.intersection(other)

    @same_types
    def __rand__(self, other):
        op = SetIntersection(self, flipped=True)
        return op(other)

    def intersection_update(self, *others):
        """Update the set, keeping only elements found in it and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            The same behavior as at :func:`difference_update` applies.
        """
        op = SetIntersection(self, update=True)
        op(*others)

    @same_types
    def __iand__(self, other):
        """Update the set, keeping only elements found in it and the *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: None

        .. note::
            The same behavior as at :func:`__isub__` applies.
        """
        op = SetIntersection(self, update=True)
        return op(other)

    def union(self, *others, **kwargs):
        """Return a new set with elements from the set and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :param return_cls: Keyword argument, type of result, defaults to
                            the same type as collection (:class:`Set`,
                            if not inherited).
        :rtype: :class:`Set` or collection of type specified in
                *return_cls* argument

        .. note::
            The same behavior as at :func:`difference` applies.
        """
        return_cls = kwargs.get('return_cls', type(self))
        op = SetUnion(self, return_cls=return_cls)
        return op(*others)

    @same_types
    def __or__(self, other):
        """Return a new set with elements from the set and the *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: type of the first operand

        .. note::
            The same behavior as at :func:`__sub__` applies.
        """
        return self.union(other)

    @same_types
    def __ror__(self, other):
        return self.union(other, return_cls=other.__class__)

    def _update(self, data, others=None, pipe=None):
        super(Set, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        others = [data] + list(others or [])
        elements = [
            self._pickle(x) for x in frozenset(itertools.chain(*others))
        ]

        redis.sadd(self.key, *elements)

    def update(self, *others):
        """Update the set, adding elements from all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            The same behavior as at :func:`difference_update` applies.
        """
        op = SetUnion(self, update=True)
        op(*others)

    @same_types
    def __ior__(self, other):
        """Update the set, adding elements from the *other*.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: None

        .. note::
            The same behavior as at :func:`__isub__` applies.
        """
        op = SetUnion(self, update=True)
        return op(other)

    def symmetric_difference(self, other, **kwargs):
        """Return a new set with elements in either the set or *other* but not
        both.

        :param others: Any kind of iterable.
        :param return_cls: Keyword argument, type of result, defaults to
                            the same type as collection (:class:`Set`,
                            if not inherited).
        :rtype: :class:`Set` or collection of type specified in
                *return_cls* argument

        .. note::
            The same behavior as at :func:`difference` applies.
        """
        return_cls = kwargs.get('return_cls', type(self))
        op = SetSymmetricDifference(self, return_cls=return_cls)
        return op(other)

    @same_types
    def __xor__(self, other):
        """Update the set, keeping only elements found in either set, but not
        in both.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: type of the first operand

        .. note::
            The same behavior as at :func:`__sub__` applies.
        """
        return self.symmetric_difference(other)

    @same_types
    def __rxor__(self, other):
        return self.symmetric_difference(other, return_cls=other.__class__)

    def symmetric_difference_update(self, other):
        """Update the set, keeping only elements found in either set, but not
        in both.

        :param others: Any kind of iterable.
        :rtype: None

        .. note::
            A bit different behavior takes place in comparing
            with the one described at :func:`difference_update`. Operation
            is **always performed in Redis**, regardless the types given.
            If *others* are instances of :class:`Set`, the performance
            should be better as no transfer of data is necessary at all.
        """
        op = SetSymmetricDifference(self, update=True)
        op(other)

    @same_types
    def __ixor__(self, other):
        """Update the set, keeping only elements found in either set, but not
        in both.

        :param other: Set object (instance of :class:`collections.Set`
                      ABC, so built-in sets and frozensets are also accepted),
                      otherwise :exc:`TypeError` is raised.
        :rtype: None

        .. note::
            The same behavior as at :func:`__isub__` applies.
        """
        op = SetSymmetricDifference(self, update=True)
        return op(other)

    def __eq__(self, other):
        if not isinstance(other, collections.Set):
            return NotImplemented
        if isinstance(other, Set):
            with self.redis.pipeline() as pipe:
                pipe.smembers(self.key)
                pipe.smembers(other.key)
                members1, members2 = pipe.execute()
            return members1 == members2
        return frozenset(self) == frozenset(other)

    def __le__(self, other):
        if not isinstance(other, collections.Set):
            return NotImplemented
        return self.issubset(other)

    def __lt__(self, other):
        if not isinstance(other, collections.Set):
            return NotImplemented
        if isinstance(other, Set):
            with self.redis.pipeline() as pipe:
                pipe.smembers(self.key)
                pipe.sinter(self.key, other.key)
                pipe.scard(other.key)
                members, inters, other_size = pipe.execute()
            return (members == inters and len(members) != other_size)
        return frozenset(self) < frozenset(other)

    def issubset(self, other):
        """Test whether every element in the set is in other.

        :param other: Any kind of iterable.
        :rtype: boolean
        """
        if isinstance(other, Set):
            with self.redis.pipeline() as pipe:
                pipe.smembers(self.key)
                pipe.sinter(self.key, other.key)
                members, inters = pipe.execute()
            return members == inters
        return frozenset(self) <= frozenset(other)

    def issuperset(self, other):
        """Test whether every element in other is in the set.

        :param other: Any kind of iterable.
        :rtype: boolean
        """
        if isinstance(other, collections.Set):
            return other <= self
        else:
            return frozenset(other) <= self

    def _repr_data(self, data):
        list_repr = repr(list(data))
        set_repr = '{' + list_repr[1:-1] + '}'
        return set_repr



# -*- coding: utf-8 -*-
"""
dicts
~~~~~

Collections based on dict interface.
"""
from __future__ import division, print_function, unicode_literals

import collections
import operator

import six

from .base import RedisCollection


class Dict(RedisCollection, collections.MutableMapping):
    """Mutable **mapping** collection aiming to have the same API as the
    standard mapping type, dict. See `dict
    <http://docs.python.org/2/library/stdtypes.html#mapping-types-dict>`_ for
    further details. The Redis implementation is based on the
    `hash <http://redis.io/commands#hash>`_ type.

    .. warning::
        In comparing with original :class:`dict` type, :class:`Dict` does not
        implement methods :func:`viewitems`, :func:`viewkeys`, and
        :func:`viewvalues`.

    .. note::
        Some operations, which are usually not used so often, can be more
        efficient than their "popular" equivalents. For example, :func:`get`
        should be preffered over the classic ``d[key]`` approach.
    """

    class __missing_value(object):
        def __repr__(self):
            return '<missing value>'  # for purposes of generated documentation
    __marker = __missing_value()

    def __init__(self, *args, **kwargs):
        """Breakes the original :class:`dict` API, because there is no support
        for keyword syntax. The only single way to create :class:`Dict`
        object is to pass iterable or mapping as the first argument.

        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str
        :param writeback: If ``True`` keep a local cache of changes for storing
                          modifications to mutable values. Changes will be
                          written to Redis after calling the ``sync`` method.
        :type key: bool

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.

        .. warning::
            As mentioned, :class:`Dict` does not support following
            initialization syntax: ``d = Dict(a=1, b=2)``
        """
        data = args[0] if args else kwargs.pop('data', None)
        writeback = kwargs.pop('writeback', False)
        super(Dict, self).__init__(*args, **kwargs)

        self.writeback = writeback
        self.cache = {}

        if data:
            self.update(data)

    def _get_hash_dict(self, key, redis):
        key_hash = hash(key)
        D = redis.hget(self.key, key_hash)
        D = {} if D is None else self._unpickle(D)

        return key_hash, D

    def __len__(self):
        """Return the number of items in the dictionary.

        .. note::
            Due to implementation on Redis side, this method is inefficient.
            The time taken is varies with the number of keys in stored.
        """
        ret = 0
        for D in six.itervalues(self.redis.hgetall(self.key)):
            ret += len(self._unpickle(D))

        return ret

    def __iter__(self, pipe=None):
        """Return an iterator over the keys of the dictionary."""
        pipe = pipe or self.redis
        for D in six.itervalues(pipe.hgetall(self.key)):
            for k in six.iterkeys(self._unpickle(D)):
                yield k

    def __contains__(self, key):
        """Return ``True`` if *key* is present, else ``False``."""
        key_hash, D = self._get_hash_dict(key, self.redis)

        return key in D

    def getmany(self, *keys):
        """Return the value for *keys*. If particular key is not in the
        dictionary, return :obj:`None`.
        """
        D_subset = {}
        for D in self.redis.hmget(self.key, *(hash(k) for k in keys)):
            if D is not None:
                D_subset.update(self._unpickle(D))

        return [D_subset.get(key) for key in keys]

    def __getitem__(self, key):
        """Return the item of dictionary with key *key*. Raises a
        :exc:`KeyError` if key is not in the map.

        If a subclass of :class:`Dict` defines a method :func:`__missing__`, if
        the key *key* is not present, the ``d[key]`` operation calls that
        method with the key *key* as argument. The ``d[key]`` operation
        then returns or raises whatever is returned or raised by
        the ``__missing__(key)`` call if the key is not present.
        """
        try:
            value = self.cache[key]
        except KeyError:
            key_hash, D = self._get_hash_dict(key, self.redis)
            try:
                value = D[key]
            except KeyError:
                if hasattr(self, '__missing__'):
                    return self.__missing__(key)
                else:
                    raise

        if self.writeback:
            self.cache[key] = value
        return value

    def __setitem__(self, key, value):
        """Set ``d[key]`` to *value*."""
        key_hash, D = self._get_hash_dict(key, self.redis)
        D[key] = value

        self.redis.hset(self.key, key_hash, self._pickle(D))
        if self.writeback:
            self.cache[key] = value

    def __delitem__(self, key):
        """Remove ``d[key]`` from dictionary.
        Raises a :func:`KeyError` if *key* is not in the map.
        """
        key_hash, D = self._get_hash_dict(key, self.redis)
        del D[key]

        if D:
            self.redis.hset(self.key, key_hash, self._pickle(D))
        else:
            self.redis.hdel(self.key, key_hash)

        try:
            del self.cache[key]
        except KeyError:
            pass

    def _data(self, pipe=None):
        """Returns a Python dictionary with the same values as this object"""
        redis = self.redis if pipe is None else pipe

        ret = {}
        for D in six.itervalues(redis.hgetall(self.key)):
            ret.update(self._unpickle(D))

        return ret

    def items(self):
        """Return a copy of the dictionary's list of ``(key, value)`` pairs."""
        return list(self.iteritems())

    def iteritems(self):
        """Return an iterator over the dictionary's ``(key, value)`` pairs."""
        for D in six.itervalues(self.redis.hgetall(self.key)):
            for k, v in six.iteritems(self._unpickle(D)):
                try:
                    yield k, self.cache[k]
                except KeyError:
                    yield k, v

    def keys(self):
        """Return a copy of the dictionary's list of keys."""
        return list(self.iterkeys())

    def iter(self):
        """Return an iterator over the keys of the dictionary.
        This is a shortcut for :func:`iterkeys()`.
        """
        return self.__iter__()

    def iterkeys(self):
        """Return an iterator over the dictionary's keys."""
        for D in six.itervalues(self.redis.hgetall(self.key)):
            for k in six.iterkeys(self._unpickle(D)):
                yield k

    def values(self):
        """Return a copy of the dictionary's list of values."""
        return list(self.itervalues())

    def itervalues(self):
        """Return an iterator over the dictionary's values."""
        for D in six.itervalues(self.redis.hgetall(self.key)):
            for k, v in six.iteritems(self._unpickle(D)):
                try:
                    yield self.cache[k]
                except KeyError:
                    yield v

    def pop(self, key, default=__marker):
        """If *key* is in the dictionary, remove it and return its value,
        else return *default*. If *default* is not given and *key* is not
        in the dictionary, a :exc:`KeyError` is raised.
        """
        def pop_trans(pipe):
            key_hash, D = self._get_hash_dict(key, pipe)
            value = D.pop(key, default)

            pipe.multi()
            if D:
                pipe.hset(self.key, key_hash, self._pickle(D))
            else:
                pipe.hdel(self.key, key_hash)

            return value

        value = self._transaction(pop_trans)
        try:
            value = self.cache[key]
        except KeyError:
            pass
        else:
            del self.cache[key]

        if value is self.__marker:
            raise KeyError(key)

        return value

    def popitem(self):
        """Remove and return an arbitrary ``(key, value)`` pair from
        the dictionary.

        :func:`popitem` is useful to destructively iterate over
        a dictionary, as often used in set algorithms. If
        the dictionary is empty, calling :func:`popitem` raises
        a :exc:`KeyError`.
        """
        def popitem_trans(pipe):
            entries = pipe.hgetall(self.key)
            if not entries:
                raise KeyError

            key_hash, D = entries.popitem()
            D = self._unpickle(D)
            item = D.popitem()

            pipe.multi()
            if D:
                pipe.hset(self.key, key_hash, self._pickle(D))
            else:
                pipe.hdel(self.key, key_hash)

            return item

        key, value = self._transaction(popitem_trans)
        try:
            item = key, self.cache[key]
        except KeyError:
            item = key, value
        else:
            del self.cache[key]

        return item

    def setdefault(self, key, default=None):
        """If *key* is in the dictionary, return its value.
        If not, insert *key* with a value of *default* and
        return *default*. *default* defaults to :obj:`None`.
        """
        def setdefault_trans(pipe):
            key_hash, D = self._get_hash_dict(key, pipe)
            value = D.setdefault(key, default)
            if value == default:
                pipe.hset(self.key, key_hash, self._pickle(D))

            return value

        try:
            value = self.cache[key]
        except KeyError:
            value = self._transaction(setdefault_trans)

        if self.writeback:
            self.cache[key] = value
        return value

    def _update_helper(self, other, use_redis=False):
        def _update_helper_trans(pipe):
            data = {}

            if use_redis:
                for D in six.itervalues(pipe.hgetall(other.key)):
                    data.update(self._unpickle(D))
            else:
                data.update(other)

            D_load = {}
            for key, value in six.iteritems(data):
                key_hash = hash(key)
                D_load.setdefault(key_hash, {})
                D_load[key_hash][key] = value

            pipe.multi()
            for key_hash, D in six.iteritems(D_load):
                pipe.hset(self.key, key_hash, self._pickle(D))

            if self.writeback:
                self.cache.update(data)

        if use_redis:
            self._transaction(_update_helper_trans, other.key)
        else:
            self._transaction(_update_helper_trans)

    def update(self, other=None, **kwargs):
        """Update the dictionary with the key/value pairs from *other*,
        overwriting existing keys. Return :obj:`None`.

        :func:`update` accepts either another dictionary object or
        an iterable of key/value pairs (as tuples or other iterables
        of length two). If keyword arguments are specified, the
        dictionary is then updated with those key/value pairs:
        ``d.update(red=1, blue=2)``.
        """
        if other is not None:
            if isinstance(other, RedisCollection):
                self._update_helper(other, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other)
            else:
                self._update_helper({k: v for k, v in other})

        if kwargs:
            self._update_helper(kwargs)

    def copy(self, key=None):
        other = self.__class__(redis=self.redis, key=key)
        other.update(self)

        return other

    def clear(self):
        self.redis.delete(self.key)
        self.cache.clear()

    @classmethod
    def fromkeys(cls, seq, value=None, **kwargs):
        """Create a new dictionary with keys from *seq* and values set to
        *value*.

        .. note::
            :func:`fromkeys` is a class method that returns a new dictionary.
            *value* defaults to :obj:`None`. It is possible to specify
            additional keyword arguments to be passed to :func:`__init__` of
            the new object.
        """
        values = ((item, value) for item in seq)
        return cls(values, **kwargs)

    def _repr_data(self, data):
        return repr(dict(data))

    def __enter__(self):
        self.writeback = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.sync()

    def sync(self):
        def sync_trans(pipe):
            D_load = {}
            for key, value in six.iteritems(self.cache):
                key_hash = hash(key)
                D_load.setdefault(key_hash, {})
                D_load[key_hash][key] = value

            pipe.multi()
            for key_hash, D in six.iteritems(D_load):
                pipe.hset(self.key, key_hash, self._pickle(D))

        self._transaction(sync_trans)
        self.cache = {}


class Counter(Dict):
    """Mutable **mapping** collection aiming to have the same API as
    :class:`collections.Counter`. See `Counter
    <http://docs.python.org/2/library/collections.html#collections.Counter>`_
    for further details. The Redis implementation is based on the
    `hash <http://redis.io/commands#hash>`_ type.

    .. warning::
        Not available in Python 2.6.

    .. warning::
        Note that this :class:`Counter` does not implement
        methods :func:`viewitems`, :func:`viewkeys`, and :func:`viewvalues`,
        which are available in Python 2.7's version.
    """

    _same_types = (collections.Counter,)

    def __init__(self, *args, **kwargs):
        """Breakes the original :class:`Counter` API, because there is no
        support for keyword syntax. The only single way to create
        :class:`Counter` object is to pass iterable or mapping as the first
        argument. Iterable is expected to be a sequence of elements,
        not a sequence of ``(key, value)`` pairs.

        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.

        .. warning::
            As mentioned, :class:`Counter` does not support following
            initialization syntax: ``c = Counter(a=1, b=2)``
        """
        super(Counter, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        return 0

    def most_common(self, n=None):
        """Return a list of the *n* most common elements and their counts
        from the most common to the least. If *n* is not specified,
        :func:`most_common` returns *all* elements in the counter.
        Elements with equal counts are ordered arbitrarily.
        """
        return collections.Counter(self).most_common(n)

    def elements(self, n=None):
        """Return an iterator over elements repeating each as many times as
        its count. Elements are returned in arbitrary order. If an element's
        count is less than one, :func:`elements` will ignore it.
        """
        return collections.Counter(self).elements()

    @classmethod
    def fromkeys(cls, iterable, v=None):
        raise NotImplementedError(
            'Counter.fromkeys() is undefined.  Use Counter(iterable) instead.'
        )

    def _update_helper(self, other, op, use_redis=False):
        def _update_helper_trans(pipe):
            data = {}

            if use_redis:
                for D in six.itervalues(pipe.hgetall(other.key)):
                    data.update(self._unpickle(D))
            else:
                data.update(other)

            D_load = {}
            for key, value in six.iteritems(data):
                key_hash = hash(key)
                D_load.setdefault(key_hash, {})
                D_load[key_hash][key] = op(self.get(key, 0), value)

            pipe.multi()
            for key_hash, D in six.iteritems(D_load):
                pipe.hset(self.key, key_hash, self._pickle(D))

            if self.writeback:
                self.cache.update(data)

        if use_redis:
            self._transaction(_update_helper_trans, other.key)
        else:
            self._transaction(_update_helper_trans)

    def update(self, other=None, **kwargs):
        """Elements are counted from an *iterable* or added-in from another
        *mapping* (or counter). Like :func:`dict.update` but adds counts
        instead of replacing them. Also, the *iterable* is expected to be
        a sequence of elements, not a sequence of ``(key, value)`` pairs.
        """
        if other is not None:
            if isinstance(other, Dict):
                self._update_helper(other, operator.add, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other, operator.add)
            else:
                self._update_helper(collections.Counter(other), operator.add)

        if kwargs:
            self._update_helper(kwargs, operator.add)

    def subtract(self, other=None, **kwargs):
        """Elements are subtracted from an *iterable* or from another
        *mapping* (or counter). Like :func:`dict.update` but subtracts
        counts instead of replacing them.
        """
        if other is not None:
            if isinstance(other, Dict):
                self._update_helper(other, operator.sub, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other, operator.sub)
            else:
                self._update_helper(collections.Counter(other), operator.sub)

        if kwargs:
            self._update_helper(kwargs, operator.sub)

    def __delitem__(self, key):
        """Like :func:`dict.__delitem__`, but does not raise KeyError for
        missing values.
        """
        try:
            super(Counter, self).__delitem__(key)
        except KeyError:
            pass

    def _op_helper(self, other, op, swap_args=False, inplace=False):
        def op_trans(pipe):
            # Get a collections.Counter copy of `self`
            self_counter = collections.Counter(self._data(pipe))

            # If `other` is also Redis-backed we'll want to pull its values
            # with the same transaction-provided pipeline as for `self`.
            if use_redis:
                other_counter = collections.Counter(other._data(pipe))
            else:
                other_counter = other

            # Unary case
            if other is None:
                result = op(self_counter)
            # Reversed case
            elif swap_args:
                result = op(other_counter, self_counter)
            # Normal case
            else:
                result = op(self_counter, other_counter)

            # If we're not updating `self`, we're finished
            if not inplace:
                return result

            # Otherwise we need to update `self` in this transaction
            D_load = {}
            for key, value in six.iteritems(result):
                key_hash = hash(key)
                D_load.setdefault(key_hash, {})
                D_load[key_hash][key] = value

            pipe.multi()
            pipe.delete(self.key)
            for key_hash, D in six.iteritems(D_load):
                pipe.hset(self.key, key_hash, self._pickle(D))

        if other is None:
            use_redis = False
            result = self._transaction(op_trans, None)
        elif isinstance(other, Counter):
            use_redis = True
            result = self._transaction(op_trans, other.key)
        elif isinstance(other, collections.Counter):
            use_redis = False
            result = self._transaction(op_trans)
        else:
            raise TypeError('Unsupported type {}'.format(type(other)))

        if inplace:
            return self
        else:
            new_instance = self.__class__(redis=self.redis)
            new_instance.update(result)
            return new_instance

    def __add__(self, other):
        return self._op_helper(other, operator.add)

    def __radd__(self, other):
        return self._op_helper(other, operator.add, swap_args=True)

    def __sub__(self, other):
        return self._op_helper(other, operator.sub)

    def __rsub__(self, other):
        return self._op_helper(other, operator.sub, swap_args=True)

    def __or__(self, other):
        return self._op_helper(other, operator.or_)

    def __ror__(self, other):
        return self._op_helper(other, operator.or_, swap_args=True)

    def __and__(self, other):
        return self._op_helper(other, operator.and_)

    def __rand__(self, other):
        return self._op_helper(other, operator.and_, swap_args=True)

    def __iadd__(self, other):
        return self._op_helper(other, operator.add, inplace=True)

    def __isub__(self, other):
        return self._op_helper(other, operator.sub, inplace=True)

    def __ior__(self, other):
        return self._op_helper(other, operator.ior, inplace=True)

    def __iand__(self, other):
        return self._op_helper(other, operator.iand, inplace=True)

    if not six.PY2:
        def __pos__(self):
            return self._op_helper(None, operator.pos)

        def __neg__(self):
            return self._op_helper(None, operator.neg)


class DefaultDict(Dict):
    """Mutable **mapping** collection aiming to have the same API as
    :class:`collections.defaultdict`. See
    `defaultdict  <https://docs.python.org/2/library/collections.html`_ for
    further details. The Redis implementation is based on the
    `hash <http://redis.io/commands#hash>`_ type.

    .. warning::
        Note that this :class:`DefaultDict` does not implement
        methods :func:`viewitems`, :func:`viewkeys`, and :func:`viewvalues`,
        which are available in Python 2.7's version.
    """

    def __init__(self, *args, **kwargs):
        """Breakes the original :class:`defaultdict` API, because there is no
        support for keyword syntax. The only single way to create
        :class:`defaultdict` object is to pass an iterable or mapping as the
        second argument.

        :param default_factory: Used to provide default values for missing
                                keys.
        :type default_factory: callable or None
        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.

        .. warning::
            As mentioned, :class:`DefaultDict` does not support following
            initialization syntax: ``d = DefaultDict(None, a=1, b=2)``
        """
        kwargs.setdefault('writeback', True)
        if args:
            default_factory = args[0]
            args = args[1:]
        else:
            default_factory = None

        super(DefaultDict, self).__init__(*args, **kwargs)

        if default_factory is None:
            pass
        elif not callable(default_factory):
            raise TypeError('first argument must be callable or None')
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        value = self.default_factory()
        self[key] = value
        return value

    def copy(self, **kwargs):
        other = self.__class__(self.default_factory, **kwargs)
        other.update(self)

        return other



# -*- coding: utf-8 -*-
"""
base
~~~~
"""
from __future__ import division, print_function, unicode_literals

import abc
import uuid
try:
    import cPickle as pickle
except ImportError:
    import pickle as pickle  # NOQA
import functools

import redis
import six


def same_types(fn):
    """Decorator, helps to check whether operands are of
    the same type as *self*. It is possible to extend
    allowed types by defining ``_same_types`` class property
    with tuple of allowed classes.
    """
    @functools.wraps(fn)
    def wrapper(self, *args):
        types = (self.__class__,) + self._same_types

        # all args must be an instance of any of the types
        allowed = all([
            any([isinstance(arg, t) for t in types])
            for arg in args
        ])

        if not allowed:
            types_msg = ', '.join(types[:-1])
            types_msg = ' or '.join([types_msg, types[-1]])
            message = ('Only instances of %s are '
                       'supported as operand types.') % types_msg
            raise TypeError(message)

        return fn(self, *args)
    return wrapper


@six.add_metaclass(abc.ABCMeta)
class RedisCollection(object):
    """Abstract class providing backend functionality for all the other
    Redis collections.
    """

    _same_types = ()

    not_impl_msg = ('Cannot be implemented efficiently or atomically '
                    'due to limitations in Redis command set.')

    @abc.abstractmethod
    def __init__(self, data=None, redis=None, key=None):
        """
        :param data: Initial data.
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key of the collection. Collections with the same key
                    point to the same data. If not provided, default random
                    string is generated.
        :type key: str

        .. note::
            :func:`uuid.uuid4` is used for default key generation.
            If you are not satisfied with its `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding
            internal method :func:`_create_key`.
        """
        #: Redis client instance. :class:`StrictRedis` object with default
        #: connection settings is used if not set by :func:`__init__`.
        self.redis = redis or self._create_redis()

        #: Redis key of the collection.
        self.key = key or self._create_key()

        # data initialization
        if data is not None:
            if isinstance(data, RedisCollection):
                # wrap into transaction
                def init_trans(pipe):
                    d = data._data(pipe=pipe)  # retrieve
                    pipe.multi()
                    self._init_data(d, pipe=pipe)  # store
                self._transaction(init_trans)
            else:
                self._init_data(data)

    def _create_new(self, data=None, key=None, pipe=None, cls=None):
        """Shorthand for creating instances of any collections. *cls*
        specifies the type of collection to be created. If subclass of
        :class:`RedisCollection` given, all settings from current ``self``
        are propagated.

        :param data: Initial data in form of a classic, built-in collection.
        :param key: Redis key of the instance. Ignored if *cls* is not a
                   :class:`RedisCollection` subclass.
        :type key: string
        :param pipe: Redis pipe in case creation is performed as a part
                     of transaction. Ignored if *cls* is not a
                     :class:`RedisCollection` subclass.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        :param cls: Type of the collection. Defaults to ``self.__class__``.
        :type cls: Class object.
        """
        assert not isinstance(data, RedisCollection), \
            "Not atomic. Use '_data()' within a transaction first."
        cls = cls or self.__class__
        if issubclass(cls, RedisCollection):
            settings = {
                'key': key,
                'redis': self.redis,
            }

            if pipe is not None and data:
                # here we cannot use cls(data, **settings), because
                # that would not be atomic within the transaction
                new = cls(**settings)
                new._init_data(data, pipe=pipe)
                return new
            return cls(data, **settings)

        return cls(data) if data else cls()

    def _init_data(self, data, pipe=None):
        """Data initialization helper.

        :param data: Initial data in form of a classic, built-in collection.
        :param pipe: Redis pipe in case creation is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """
        assert not isinstance(data, RedisCollection), \
            "Not atomic. Use '_data()' within a transaction first."

        # if not in pipe, make your own
        p = pipe if pipe is not None else self.redis.pipeline()
        if data is not None:
            self._clear(pipe=p)
            if data:
                # non-empty data, populate collection
                self._update(data, pipe=p)
        if pipe is None:
            # own pipe, execute it
            p.execute()

    def _create_redis(self):
        """Creates default Redis connection.

        :rtype: :class:`redis.StrictRedis`
        """
        return redis.StrictRedis()

    def _create_key(self):
        """Creates new Redis key.

        :rtype: string

        .. note::
            :func:`uuid.uuid4` is used. If you are not satisfied with its
            `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding this
            method.
        """
        return uuid.uuid4().hex

    @abc.abstractmethod
    def _data(self, pipe=None):
        """Helper for getting collection's data within a transaction.

        :param pipe: Redis pipe in case creation is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """
        pass

    def _pickle(self, data):
        """Converts given data to string.

        :param data: Data to be serialized.
        :type data: anything serializable
        :rtype: string
        """
        return pickle.dumps(data)

    def _unpickle(self, string):
        """Converts given string serialization back to corresponding data.
        If :obj:`None` or empty string given, :obj:`None` is returned.

        :param string: String to be unserialized.
        :type string: string
        :rtype: anything serializable
        """
        return pickle.loads(string) if string else None

    def _update(self, data, pipe=None):
        """Helper for update operations.

        :param data: Data for update in form of a classic, built-in collection.
        :param pipe: Redis pipe in case update is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """
        assert not isinstance(data, RedisCollection), \
            "Not atomic. Use '_data()' within a transaction first."

    def _clear(self, pipe=None):
        """Helper for clear operations.

        :param pipe: Redis pipe in case update is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """
        redis = pipe if pipe is not None else self.redis
        redis.delete(self.key)

    def clear(self):
        """Completely cleares the collection's data."""
        self._clear()

    def _transaction(self, fn, *extra_keys):
        """Helper simplifying code within watched transaction.

        Takes *fn*, function treated as a transaction. Returns whatever
        *fn* returns. ``self.key`` is watched. *fn* takes *pipe* as the
        only argument.

        :param fn: Closure treated as a transaction.
        :type fn: function *fn(pipe)*
        :param extra_keys: Optional list of additional keys to watch.
        :type extra_keys: list
        :rtype: whatever *fn* returns
        """
        results = []

        def trans(pipe):
            results.append(fn(pipe))

        self.redis.transaction(trans, self.key, *extra_keys)
        return results[0]

    def copy(self, key=None):
        """Return a copy of the collection.

        :param key: Key of the new collection. Defaults to auto-generated.
        :type key: string
        """
        def copy_trans(pipe):
            data = self._data(pipe=pipe)  # retrieve
            pipe.multi()
            return self._create_new(data, key=key, pipe=pipe)  # store
        return self._transaction(copy_trans, key)

    def _repr_data(self, data):
        return repr(data)

    def __repr__(self):
        cls_name = self.__class__.__name__
        data = self._repr_data(self._data())
        return '<redis_collections.%s at %s %s>' % (cls_name, self.key, data)



# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

import os
import re
import subprocess
import sys

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages  # NOQA

# Hack to prevent stupid "TypeError: 'NoneType' object is not callable"
# error in multiprocessing/util.py _exit_function when running `python
# setup.py test`
try:
    import multiprocessing  # NOQA
except ImportError:
    pass


base_path = os.path.dirname(__file__)


# version
meta_file = os.path.join(base_path, 'redis_collections/__init__.py')
meta_file_contents = open(meta_file).read()
meta = dict(re.findall(r'__([^_]+)__ = \'([^\']*)\'', meta_file_contents))


# release a version, publish to GitHub and PyPI
if sys.argv[-1] == 'publish':
    subprocess.check_call(['git', 'tag', 'v' + meta['version']])
    subprocess.check_call(['git', 'push', '--tags', 'origin', 'master:master'])
    # ...the rest happens on Travis CI
    sys.exit()


setup(
    name=meta['title'],
    version=meta['version'],
    description='Set of basic Python collections backed by Redis.',
    long_description=open('README.rst').read(),
    author=meta['author'],
    author_email='mail@honzajavorek.cz',
    url='https://github.com/honzajavorek/redis-collections',
    license=open('LICENSE').read(),
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=['redis>=2.7.2', 'six>=1.10.0'],
    zip_safe=False,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Database',
    )
)
